import os
import re
import tempfile
from typing import Optional
import uuid
from fastapi import HTTPException
import yaml

from rdkit import Chem
from rdkit.Chem import RegistrationHash
from rdkit.Chem.MolStandardize import rdMolStandardize
from rdkit.Chem.RegistrationHash import HashLayer

from app import models
from sqlalchemy.orm import Session


class MoleculeStandardizationConfig:
    def __init__(self, db: Optional[Session] = None):
        self._config = None
        self.db = db

    @property
    def config(self) -> dict:
        if self._config is None:
            setting = (
                self.db.query(models.Settings).filter(models.Settings.name == "Molecule standardization rules").first()
            )
            if not setting:
                raise Exception("Molecule standardization config not found in settings table.")
            self._config = yaml.safe_load(setting.value)
        return self._config

    def clear_cache(self):
        self._config = None

    def set_db(self, db: Session):
        self.db = db


molecule_standardization_config = MoleculeStandardizationConfig()


def get_molecule_standardization_config(db: Optional[Session] = None):
    global molecule_standardization_config
    if db is not None:
        molecule_standardization_config.set_db(db)
    return molecule_standardization_config


def standardize_mol(mol: Chem.Mol, db: Optional[Session] = None) -> Chem.Mol:
    """
    Standardizes a given RDKit molecule using operations defined in the
    molecule standardization settings.

    The operations are dynamically executed in the order defined in the setting, but only if they are enabled.

    Args:
        mol (Chem.Mol): The molecule to standardize.

    Returns:
        Chem.Mol: The standardized molecule after performing all configured operations.
    """
    config = get_molecule_standardization_config(db).config
    # Apply only the enabled operations in the order of declaration in the config.
    for operation in config.get("operations", []):
        operation_type = operation.get("type")
        is_enabled = operation.get("enable", True)

        if not is_enabled:
            continue

        if not operation_type:
            raise ValueError("Operation type is missing in the configuration.")

        mol = apply_standardizer_operation(mol, operation_type)

    return mol


def apply_standardizer_operation(mol: Chem.Mol, operation_type: str) -> Chem.Mol:
    """
    Applies a specific operation to the molecule based on the operation type.

    Args:
        mol (Chem.Mol): The molecule to modify.
        operation_type (str): The type of standardization operation to perform.

    Returns:
        Chem.Mol: The transformed molecule.
    """
    operation_map = {
        "Cleanup": rdMolStandardize.Cleanup,
        "FragmentParent": rdMolStandardize.FragmentParent,
        "RemoveHs": Chem.RemoveHs,
        "Uncharger": lambda mol: rdMolStandardize.Uncharger().uncharge(mol),
    }

    if operation_type not in operation_map:
        raise ValueError(f"Unknown operation type: {operation_type}")

    return operation_map[operation_type](mol)


def generate_hash_layers(mol: Chem.Mol) -> dict:
    """
    Generate layers for a given molecule.

    This function calculates the layers using the `RegistrationHash` module.

    Args:
        mol: An RDKit molecule object (`rdkit.Chem.Mol`) for which the layers
                  will be generated.

    Returns:
        dict: A dictionary containing the layers used to compute the MolHash.
    """

    return RegistrationHash.GetMolLayers(mol, enable_tautomer_hash_v2=True)


def generate_uuid_from_string(input_string: str) -> uuid.UUID:
    """
    Generate a UUID hash for a given input string, for hashing different molecule layers.

    Args:
        input_string (str): The input string to hash.

    Returns:
        uuid.UUID: The UUID hash of the input string, ready for PostgreSQL UUID type.
    """
    return uuid.uuid5(uuid.NAMESPACE_DNS, input_string)


def calculate_tautomer_hash(mol: Chem.Mol) -> str:
    """
    Calculate the tautomer hash for a given molecule.
    """
    return generate_uuid_from_string(generate_hash_layers(mol)[HashLayer.TAUTOMER_HASH])


def calculate_no_stereo_smiles_hash(mol: Chem.Mol) -> str:
    """Calculate the no-stereo SMILES hash for a given molecule."""

    return generate_uuid_from_string(generate_hash_layers(mol)[HashLayer.NO_STEREO_SMILES])


def calculate_no_stereo_tautomer_hash(mol: Chem.Mol) -> str:
    """
    Calculate the no-stereo tautomer hash for a given molecule.
    """
    return generate_uuid_from_string(generate_hash_layers(mol)[HashLayer.NO_STEREO_TAUTOMER_HASH])


def validate_rdkit_call(func, *args, err_msg_prefix: str = ""):
    """Run an RDKit function while capturing stderr output and handling errors.

    Args:
        func: RDKit callable (e.g., Chem.MolFromSmiles, Chem.MolToInchi)
        *args: Arguments to pass to the function
        err_msg_prefix: Prefix to add to the error message
    """
    with tempfile.NamedTemporaryFile(mode="w+", delete=True) as tmp_file:
        stderr_fd = 2
        stderr_backup = os.dup(stderr_fd)

        try:
            os.dup2(tmp_file.fileno(), stderr_fd)
            result = func(*args)
        except Exception as e:
            os.dup2(stderr_backup, stderr_fd)
            os.close(stderr_backup)
            raise HTTPException(status_code=400, detail=f"{err_msg_prefix} {e}")
        finally:
            os.dup2(stderr_backup, stderr_fd)
            os.close(stderr_backup)

        tmp_file.seek(0)
        error_msg = re.search(r"ERROR:.*", tmp_file.read().strip())

    if result is None or error_msg:
        msg = f"{err_msg_prefix} {error_msg.group(0) if error_msg else 'Execution failed'}"
        raise HTTPException(status_code=400, detail=msg)
    return result



from typing import Any, Dict, List
from sqlalchemy.orm import Session
import app.models as models
from app.services.search.field_resolver import FieldResolver, FieldResolutionError
from app.services.search.query_builder import QueryBuilder, QueryBuildError
from sqlalchemy import text
from app.setup.database import DB_SCHEMA
from app.services.search.operators import SearchOperators
from app.services.search.utils.helper_functions import sanitize_field_name, prepare_search_output
from app.models import Level


class SearchEngineError(Exception):
    """Custom exception for search engine errors"""

    pass


class SearchEngine:
    """Search orchestration engine"""

    def __init__(self, db: Session):
        self.db = db
        self.db_schema = DB_SCHEMA
        self.field_resolver = FieldResolver(self.db_schema, db)
        self.query_builder = QueryBuilder(self.field_resolver)
        self.results = None

    def search(self, request: models.SearchRequest):
        """
        Executes search request

        Args:
            request: SearchRequest object with search parameters

        Returns:
            SearchResponse with results and metadata
        """
        try:
            # Validate the request
            validation_errors = self.validate_request(request)
            if validation_errors:
                raise SearchEngineError(f"Request validation failed: {'; '.join(validation_errors)}")

            # Extract column names from output fields
            columns = [field.lower() for field in request.output]
            self.output_aliases = {sanitize_field_name(field): field for field in columns}

            # Build the SQL query
            query_info = self.query_builder.build_query(request)

            # Execute main query
            results, headers = self._execute_main_query(query_info["sql"], query_info["params"])

            return prepare_search_output(results, headers, request.output_format)

        except (FieldResolutionError, QueryBuildError) as e:
            raise SearchEngineError(f"Search execution error: {str(e)}")
        except Exception as e:
            raise SearchEngineError(f"Error: {str(e)}")

    def validate_request(self, request: models.SearchRequest) -> List[str]:
        """
        Validates search request constraints

        Returns:
            List of validation error messages
        """
        errors = []

        # Validate output fields
        for field_path in request.output:
            if not self.field_resolver.validate_field_path(field_path, request.level):
                errors.append(f"Invalid output field: {field_path}")

        # Validate filter conditions if present
        if request.filter:
            filter_errors = self._validate_filter(request.filter, request.level)
            errors.extend(filter_errors)

        return errors

    def _validate_filter(self, filter_obj: models.Filter, level: Level, path: str = "filter") -> List[str]:
        """Recursively validate filter conditions"""
        errors = []

        if isinstance(filter_obj, models.AtomicCondition):
            # Validate single atomic condition
            if not self.field_resolver.validate_field_path(filter_obj.field):
                errors.append(f"Invalid field at {path}: {filter_obj.field}")
            SearchOperators.validate_operands(filter_obj.operator, filter_obj.field)
            SearchOperators.validate_operator_value(filter_obj.operator, filter_obj.value, filter_obj.threshold)

        elif isinstance(filter_obj, models.LogicalNode):
            # Validate logical node with multiple conditions
            for i, condition in enumerate(filter_obj.conditions):
                condition_path = f"{path}.conditions[{i}]"

                if isinstance(condition, models.AtomicCondition):
                    # Validate field path
                    if not self.field_resolver.validate_field_path(condition.field):
                        errors.append(f"Invalid field at {condition_path}: {condition.field}")
                    SearchOperators.validate_operands(condition.operator, condition.field)
                    SearchOperators.validate_operator_value(condition.operator, condition.value, condition.threshold)

                elif isinstance(condition, models.LogicalNode):
                    # Recursively validate nested filters
                    nested_errors = self._validate_filter(condition, level, condition_path)
                    errors.extend(nested_errors)

        return errors

    def _execute_main_query(self, sql: str, params: Dict[str, Any]) -> List[Dict[str, Any]]:
        """Execute main query and return results as list of dictionaries"""
        try:
            # SQLAlchemy expects parameters to be passed as keyword arguments
            result = self.db.execute(text(sql), params)

            # Get column names from result
            if result.returns_rows:
                headers = [self.output_aliases[item] for item in list(result.keys())]
                rows = result.fetchall()
                return rows, headers
            else:
                return [], []

        except Exception as e:
            raise SearchEngineError(f"Main query execution failed: {str(e)}")



from typing import List, Dict, Any
from psycopg2.extensions import adapt

column_types = {
    "value_datetime": "timestamptz",
    "value_num": "double precision",
    "value_uuid": "uuid",
    "value_string": "text",
    "value_bool": "boolean",
    "value_qualifier": "smallint",
}


def values_sql(data: List[Dict[str, Any]], columns: List[str]) -> str:
    def escape_val(val, col_name):
        if val is None:
            sql_type = column_types.get(col_name)
            if sql_type:
                return f"NULL::{sql_type}"
            return "NULL"
        return adapt(val).getquoted().decode()

    rows = []
    for row in data:
        values = [escape_val(row.get(col), col) for col in columns]
        rows.append(f"({', '.join(values)})")
    return ",\n".join(rows)


def prepare_sql_parts(records: List[Dict[str, Any]]):
    cols = list(records[0].keys())
    key, *cols_without_key = cols
    values_to_sql = values_sql(records, cols)
    return cols_without_key, values_to_sql


def generate_sql(*sql_parts: str, terminate_with_select: bool = True) -> str:
    filtered_parts = [part.strip() for part in sql_parts if part and part.strip()]
    if not filtered_parts:
        return ""
    combined_sql = ",\n".join(filtered_parts)
    combined_sql += "\nSELECT 1;" if terminate_with_select else ";"
    return combined_sql


def chunked(lst, size):
    for i in range(0, len(lst), size):
        yield lst[i : i + size]



# Import common data from conftest.py (fixtures are automatically available)
from typing import List
import pytest
import pandas as pd
import io
import json
from app.utils import enums


def upload_compounds(client, smiles_list: List[str]) -> None:
    """Uploads a list of SMILES using the /v1/compounds/ endpoint."""
    response = client.get("/v1/compounds/")
    assert response.status_code == 200
    existing = response.json()
    max_id = existing[0]["id"] if existing else 0

    df = pd.DataFrame({"smiles": smiles_list})
    csv_bytes = df.to_csv(index=False).encode("utf-8")
    csv_file = io.BytesIO(csv_bytes)

    mapping = {"smiles": "smiles"}
    files = {"csv_file": ("compounds.csv", csv_file, "text/csv")}
    data = {
        "error_handling": enums.ErrorHandlingOptions.reject_row.value,
        "mapping": json.dumps(mapping),
    }

    response = client.post("/v1/compounds/", files=files, data=data)
    assert response.status_code == 200, f"Failed to upload compounds: {response.text}"

    first_new_id = max_id + 1
    return list(range(first_new_id, first_new_id + len(smiles_list)))


@pytest.mark.skip(reason="This needs to be corrected to work with new search. Old search is no longer available")
def test_create_compound_and_get_hash(client):
    """Test creating a compound and retrieving its hash using the exact search endpoint"""
    smiles = ["C[C@@](F)(Cl)c1cc2ccc[nH]c-2n1"]

    # Step 1: Create the compound
    upload_compounds(client, smiles)

    # Step 2: Use the /v1/search/compounds/exact endpoint to retrieve the hash_mol
    search_payload = {"query_smiles": smiles[0]}
    search_response = client.post("/v1/search/compounds/exact", json=search_payload)
    assert search_response.status_code == 200, f"Unexpected status code: {search_response.status_code}"
    search_results = search_response.json()

    # Verify the hash_mol matches the expected value
    assert len(search_results) == 1, "Expected exactly one result"
    mol_hash = search_results[0]["hash_mol"]
    expected_hash = "9871427f8720e3b4b3219964869a6301377f6908"
    assert mol_hash == expected_hash, f"Expected hash {expected_hash}, got {mol_hash}"


@pytest.fixture()
def predefined_compounds(client):
    """Fixture to insert predefined compounds into the database."""
    smiles_list = [
        "C[C@@](F)(Cl)c1cc2ccc[nH]c-2n1",  # Tautomer1 R
        "C[C@@](F)(Cl)c1cc2cccnc2[nH]1",  # Tautomer1 S
        "C[C@](F)(Cl)c1cc2ccc[nH]c-2n1",  # Tautomer2 R
        "C[C@](F)(Cl)c1cc2cccnc2[nH]1",  # Tautomer2 S
    ]

    compound_ids = upload_compounds(client, smiles_list)
    yield compound_ids


@pytest.mark.skip(reason="This needs to be corrected to work with new search. Old search is no longer available")
def test_search_compound_structure_tautomer(client, predefined_compounds):
    """Test tautomer search using the /v1/search/compounds/structure endpoint"""
    smiles_list = [
        "C[C@@](F)(Cl)c1cc2ccc[nH]c-2n1",  # Tautomer1 R
        "C[C@@](F)(Cl)c1cc2cccnc2[nH]1",  # Tautomer1 S
        "C[C@](F)(Cl)c1cc2ccc[nH]c-2n1",  # Tautomer2 R
        "C[C@](F)(Cl)c1cc2cccnc2[nH]1",  # Tautomer2 S
    ]

    search_payload = {
        "search_type": "tautomer",
        "query_smiles": smiles_list[0],
        "search_parameters": {},
    }
    search_response = client.post("/v1/search/compounds/structure", json=search_payload)
    assert search_response.status_code == 200, f"Unexpected status code: {search_response.status_code}"
    search_results = search_response.json()
    result_ids = [compound["id"] for compound in search_results]

    assert predefined_compounds[0] in result_ids, "Tautomer1 R should match the query"
    assert predefined_compounds[1] in result_ids, "Tautomer1 S should match the query"
    assert predefined_compounds[2] not in result_ids, "Tautomer2 R should NOT match the query"
    assert predefined_compounds[3] not in result_ids, "Tautomer2 S should NOT match the query"


@pytest.mark.skip(reason="This needs to be corrected to work with new search. Old search is no longer available")
def test_search_compound_structure_stereo(client, predefined_compounds):
    """Test stereo search using the /v1/search/compounds/structure endpoint"""
    smiles_list = [
        "C[C@@](F)(Cl)c1cc2ccc[nH]c-2n1",  # Tautomer1 R
        "C[C@@](F)(Cl)c1cc2cccnc2[nH]1",  # Tautomer1 S
        "C[C@](F)(Cl)c1cc2ccc[nH]c-2n1",  # Tautomer2 R
        "C[C@](F)(Cl)c1cc2cccnc2[nH]1",  # Tautomer2 S
    ]

    search_payload = {
        "search_type": "stereo",
        "query_smiles": smiles_list[0],
        "search_parameters": {},
    }
    search_response = client.post("/v1/search/compounds/structure", json=search_payload)
    assert search_response.status_code == 200, f"Unexpected status code: {search_response.status_code}"
    search_results = search_response.json()
    result_ids = [compound["id"] for compound in search_results]

    assert predefined_compounds[0] in result_ids, "Tautomer1 R should match the query"
    assert predefined_compounds[2] in result_ids, "Tautomer2 R should match the query"
    assert predefined_compounds[1] not in result_ids, "Tautomer1 S should NOT match the query"
    assert predefined_compounds[3] not in result_ids, "Tautomer2 S should NOT match the query"


@pytest.mark.skip(reason="This needs to be corrected to work with new search. Old search is no longer available")
def test_search_compound_structure_connectivity(client, predefined_compounds):
    """Test connectivity search using the /v1/search/compounds/structure endpoint"""
    smiles_list = [
        "C[C@@](F)(Cl)c1cc2ccc[nH]c-2n1",  # Tautomer1 R
        "C[C@@](F)(Cl)c1cc2cccnc2[nH]1",  # Tautomer1 S
        "C[C@](F)(Cl)c1cc2ccc[nH]c-2n1",  # Tautomer2 R
        "C[C@](F)(Cl)c1cc2cccnc2[nH]1",  # Tautomer2 S
    ]

    search_payload = {
        "search_type": "connectivity",
        "query_smiles": smiles_list[0],
        "search_parameters": {},
    }
    search_response = client.post("/v1/search/compounds/structure", json=search_payload)
    assert search_response.status_code == 200, f"Unexpected status code: {search_response.status_code}"
    search_results = search_response.json()
    result_ids = [compound["id"] for compound in search_results]

    assert predefined_compounds[0] in result_ids, "Tautomer1 R should match the query"
    assert predefined_compounds[1] in result_ids, "Tautomer1 S should match the query"
    assert predefined_compounds[2] in result_ids, "Tautomer2 R should match the query"
    assert predefined_compounds[3] in result_ids, "Tautomer2 S should match the query"



from tests.conftest import DATA_DIR
import pandas as pd
from app.utils import enums


def test_get_all_additions(client, preload_additions):
    response = client.get("/v1/additions/")
    assert response.status_code == 200
    additions = response.json()
    assert isinstance(additions, list)

    expected_names = pd.read_csv(DATA_DIR / "additions.csv")["name"].tolist()
    actual_names = [item["name"] for item in additions]
    assert sorted(actual_names) == sorted(expected_names)


def test_get_salts(client, preload_additions):
    response = client.get("/v1/additions/salts")
    assert response.status_code == 200
    salts = response.json()
    assert isinstance(salts, list)
    for item in salts:
        item = client.get(f"/v1/additions/{item['id']}").json()
        assert item["role"] == enums.AdditionsRole.SALT.value
    assert len(salts) == 20


def test_get_solvates(client, preload_additions):
    response = client.get("/v1/additions/solvates")
    assert response.status_code == 200
    solvates = response.json()
    assert isinstance(solvates, list)
    for item in solvates:
        item = client.get(f"/v1/additions/{item['id']}").json()
        assert item["role"] == enums.AdditionsRole.SOLVATE.value
    assert len(solvates) == 10



import enum


class CaseInsensitiveEnum(str, enum.Enum):
    @classmethod
    def _missing_(cls, value):
        if isinstance(value, str):
            for member in cls:
                if member.value.lower() == value.lower():
                    return member
        return None


class ValueType(str, enum.Enum):
    int = "int"
    double = "double"
    bool = "bool"
    datetime = "datetime"
    string = "string"
    uuid = "uuid"


class PropertyClass(CaseInsensitiveEnum):
    DECLARED = "DECLARED"
    CALCULATED = "CALCULATED"
    MEASURED = "MEASURED"
    PREDICTED = "PREDICTED"


class EntityType(CaseInsensitiveEnum):
    BATCH = "BATCH"
    COMPOUND = "COMPOUND"
    ASSAY = "ASSAY"
    ASSAY_RUN = "ASSAY_RUN"
    ASSAY_RESULT = "ASSAY_RESULT"
    SYSTEM = "SYSTEM"


class SearchEntityType(str, enum.Enum):
    BATCHES = "batches"
    COMPOUNDS = "compounds"
    ASSAYS = "assays"
    ASSAY_RUNS = "assay_runs"
    ASSAY_RESULTS = "assay_results"


class EntityTypeReduced(CaseInsensitiveEnum):
    BATCH = "BATCH"
    COMPOUND = "COMPOUND"


class AdditionsRole(CaseInsensitiveEnum):
    SALT = "SALT"
    SOLVATE = "SOLVATE"


class SynonymLevel(CaseInsensitiveEnum):
    BATCH = "BATCH"
    COMPOUND = "COMPOUND"


class ErrorHandlingOptions(str, enum.Enum):
    reject_all = "reject_all"
    reject_row = "reject_row"


class OutputFormat(str, enum.Enum):
    json = "json"
    csv = "csv"


class SearchOutputFormat(str, enum.Enum):
    json = "json"
    csv = "csv"
    parquet = "parquet"


class CompoundMatchingRule(CaseInsensitiveEnum):
    ALL_LAYERS = "ALL_LAYERS"
    STEREO_INSENSITIVE_LAYERS = "STEREO_INSENSITIVE_LAYERS"
    TAUTOMER_INSENSITIVE_LAYERS = "TAUTOMER_INSENSITIVE_LAYERS"


class LogicOp(str, enum.Enum):
    """Logical operators for combining conditions"""

    AND = "AND"
    OR = "OR"


class CompareOp(str, enum.Enum):
    """Comparison operators for atomic conditions"""

    # String operators
    EQUALS = "="
    NOT_EQUALS = "!="
    IN = "IN"
    STARTS_WITH = "STARTS WITH"
    ENDS_WITH = "ENDS WITH"
    LIKE = "LIKE"
    CONTAINS = "CONTAINS"

    # Numeric operators
    LESS_THAN = "<"
    GREATER_THAN = ">"
    LESS_THAN_OR_EQUAL = "<="
    GREATER_THAN_OR_EQUAL = ">="
    RANGE = "RANGE"

    # Datetime operators
    BEFORE = "BEFORE"
    AFTER = "AFTER"
    ON = "ON"

    # Molecular operators (RDKit)
    IS_SIMILAR = "IS SIMILAR"
    IS_SUBSTRUCTURE_OF = "IS SUBSTRUCTURE OF"
    HAS_SUBSTRUCTURE = "HAS SUBSTRUCTURE"


class OperatorType(enum.Enum):
    """Types of operators for different data types"""

    STRING = "string"
    NUMERIC = "numeric"
    DATETIME = "datetime"
    BOOLEAN = "boolean"
    MOLECULAR = "molecular"


class SettingName(str, enum.Enum):
    COMPOUND_MATCHING_RULE = "COMPOUND_MATCHING_RULE"
    COMPOUND_SEQUENCE_START = "COMPOUND_SEQUENCE_START"
    BATCH_SEQUENCE_START = "BATCH_SEQUENCE_START"
    CORPORATE_COMPOUND_ID_PATTERN = "CORPORATE_COMPOUND_ID_PATTERN"
    CORPORATE_BATCH_ID_PATTERN = "CORPORATE_BATCH_ID_PATTERN"



import pytest
from app.utils import enums
from tests.conftest import BLACK_DIR, read_json, _preload_compounds


def extract_name_entity_type(items):
    return sorted(
        [{"name": item["name"], "entity_type": item["entity_type"].upper()} for item in items], key=lambda x: x["name"]
    )


def assert_name_entity_type_equal(actual, expected):
    actual_name_entity_type = extract_name_entity_type(actual)
    expected_name_entity_type = extract_name_entity_type(expected)
    assert actual_name_entity_type == expected_name_entity_type


@pytest.mark.parametrize(
    "endpoint,schema_file,response_key,expected_keys",
    [
        ("/v1/schema/compounds", "compounds_schema.json", None, ["properties", "synonym_types"]),
        ("/v1/schema/compounds/synonyms", "compounds_schema.json", None, ["synonym_types"]),
        ("/v1/schema/batches", "batches_schema.json", "properties", ["properties", "synonym_types"]),
        ("/v1/schema/batches/synonyms", "batches_schema.json", "synonym_types", ["synonym_types"]),
    ],
)
def test_schema(client, endpoint, schema_file, response_key, expected_keys, preload_schema):
    response = client.get(endpoint)
    assert response.status_code == 200

    response_data = response.json()
    actual = response_data if response_key is None else response_data[response_key]
    expected_data = read_json(BLACK_DIR / schema_file)
    expected = []
    for key in expected_keys:
        expected.extend(expected_data.get(key, []))

    if "/compounds" in endpoint:
        expected.append({"name": "corporate_compound_id", "entity_type": enums.EntityType.COMPOUND})
    if "/batches" in endpoint:
        expected.append({"name": "corporate_batch_id", "entity_type": enums.EntityType.BATCH})

    assert_name_entity_type_equal(actual, expected)

    if "/v1/schema/batches" in endpoint:
        assert "additions" in response_data


def test_register_compounds_without_mapping(client, preload_schema):
    register_response = _preload_compounds(client, BLACK_DIR / "compounds.csv")
    assert register_response.status_code == 200

    register_data = register_response.json()
    assert len(register_data) == 54

    get_response = client.get("/v1/compounds/")
    assert get_response.status_code == 200
    compounds = get_response.json()

    def assert_properties(compound, expected_props, index):
        properties = compound.get("properties", [])
        assert len(properties) == len(expected_props), (
            f"[Compound {index}] Expected {len(expected_props)} properties, got {len(properties)}"
        )
        names = {p["name"] for p in properties}
        assert names == expected_props, f"[Compound {index}] Property names mismatch: {names} != {expected_props}"

    assert_properties(compounds[0], {"epa_compound_id", "corporate_compound_id", "MolLogP"}, index=0)
    assert_properties(compounds[8], {"epa_compound_id", "corporate_compound_id"}, index=8)


@pytest.mark.skip(reason="No test datasets contain invalid records to validate 'reject all' behaviour.")
def test_register_compounds_reject_all(client, preload_schema):
    response = _preload_compounds(
        client, BLACK_DIR / "compounds.csv", BLACK_DIR / "compounds_mapping.json", enums.ErrorHandlingOptions.reject_all
    )
    assert response.status_code == 400

    result = response.json()["detail"]
    assert result["status"] == "Success"

    data = result["data"]
    assert len(data) == 54

    item8 = data[8]
    assert item8["registration_status"] == "failed"
    assert item8["registration_error_message"] == "400: Invalid SMILES string"

    for item in data[9:]:
        assert item["registration_status"] == "not_processed"
        assert item["registration_error_message"] is None


@pytest.mark.skip(reason="No test datasets contain invalid records to validate 'reject row' behaviour.")
def test_register_compounds_reject_row(client, preload_schema):
    response = _preload_compounds(client, BLACK_DIR / "compounds.csv", BLACK_DIR / "compounds_mapping.json")
    assert response.status_code == 200

    result = response.json()
    data = result["data"]
    assert isinstance(data, list)
    assert len(data) == 54

    item8 = data[8]
    assert item8["registration_status"] == "failed"
    assert item8["registration_error_message"] == "400: Invalid SMILES string"

    item9 = data[9]
    assert item9["registration_status"] == "success"
    assert item9["registration_error_message"] is None


def test_get_compounds_list(client, preload_schema, preload_compounds):
    response = client.get("/v1/compounds/")
    assert response.status_code == 200

    result = response.json()
    assert isinstance(result, list)
    assert len(result) == 54

    first = result[0]
    assert first["id"] == 1
    assert first["canonical_smiles"] == "O=C(O)c1cccc(C(=O)O)c1"
    assert first["inchikey"] == "QQVIHTHCMHWDBS-UHFFFAOYSA-N"

    props = {p["name"]: p for p in first["properties"]}
    assert props["epa_compound_id"]["value_string"] == "EPA-001"
    assert props["cas"]["value_string"] == "121-91-5"
    assert props["common_name"]["value_string"].strip() == "1,3-Benzenedicarboxylic acid"
    assert abs(props["MolLogP"]["value_num"] - 1.083) < 1e-3


def test_get_compound_by_id(client, preload_schema, preload_compounds):
    response = client.get("/v1/compounds/2")
    assert response.status_code == 200

    result = response.json()
    assert result["id"] == 2
    assert result["canonical_smiles"] == "Oc1ccc(C(c2ccc(O)cc2)(C(F)(F)F)C(F)(F)F)cc1"
    assert result["inchikey"] == "ZFVMWEVVKGLCIJ-UHFFFAOYSA-N"
    assert result["inchi"] == (
        "InChI=1S/C15H10F6O2/c16-14(17,18)13(15(19,20)21,9-1-5-11(22)6-2-9)10-3-7-12(23)8-4-10/h1-8,22-23H"
    )
    assert result["is_archived"] is False
    assert result["batches"] == []

    props = {p["name"]: p for p in result["properties"]}

    assert props["epa_compound_id"]["value_string"] == "EPA-002"
    assert props["cas"]["value_string"] == "1478-61-1"
    assert props["common_name"]["value_string"].strip() == "Bisphenol AF"
    assert abs(props["MolLogP"]["value_num"] - 4.5085) < 1e-3



from typing import Any, Dict, List, Optional

from fastapi import HTTPException
from sqlalchemy.orm import Session
from sqlalchemy import and_, func, or_, select

from app import models
from app.utils.admin_utils import admin
from app.utils import enums, sql_utils
from app.services.registrars.base_registrar import BaseRegistrar


class AssayResultsRegistrar(BaseRegistrar):
    def __init__(self, db: Session, mapping: Optional[Dict[str, str]], error_handling: str):
        super().__init__(db, mapping, error_handling)
        self.assay_results_to_insert = []

    def _check_single_result(self, results: list, error_context: str):
        if len(results) == 0:
            raise HTTPException(status_code=400, detail=f"No {error_context} found matching the provided details")
        if len(results) > 1:
            raise HTTPException(status_code=400, detail=f"Multiple {error_context} found matching the provided details")
        return results[0]

    def _lookup_by_details(
        self,
        details: Dict[str, Any],
        details_model,
        parent_id_field: str,
        entity_type: str,
        parent_id_value: Optional[int],
    ):
        """
        Generic lookup for detail-based models, e.g. batch_details or assay_run_details.

        :param details: dict of property_name -> value
        :param details_model: model (BatchDetail or AssayRunDetail)
        :param parent_id_field: str, e.g. 'batch_id' or 'assay_run_id'
        :param parent_id_value: int or None, the parent entity id to filter details by;
                                if None, don't filter by parent id (optional)
        :return: subquery returning matching parent ids
        """
        if not details:
            raise HTTPException(status_code=400, detail="No details provided for lookup")

        property_values = []
        for prop_name, value in details.items():
            if value in (None, "", []):
                continue  # Skip this property if value is empty (we are not inserting empty values)
            prop_info = self.property_service.get_property_info(prop_name, entity_type)

            property_values.append(
                {
                    "property_id": getattr(prop_info["property"], "id"),
                    "value_column_name": prop_info["field_name"],
                    "value": value,
                }
            )

        # Construct OR conditions, where each condition ensures that a row matches a specific property and value
        num_details = len(property_values)
        or_conditions = []
        for pv in property_values:
            col = getattr(details_model, pv["value_column_name"])
            conditions = [
                details_model.property_id == pv["property_id"],
                col == pv["value"],
            ]
            if parent_id_value is not None:
                conditions.append(getattr(details_model, parent_id_field) == parent_id_value)
            cond = and_(*conditions)
            or_conditions.append(cond)

        subq = (
            self.db.query(getattr(details_model, parent_id_field))
            .filter(or_(*or_conditions))  # Match any of the provided pairs
            .group_by(getattr(details_model, parent_id_field))  # Group by parent id
            .having(
                func.count(getattr(details_model, parent_id_field)) == num_details
            )  # Keep only those with all matches
            .subquery()
        )
        return subq

    def _lookup_batch_by_details(self, batch_details: Dict[str, Any]) -> models.Batch:
        subq = self._lookup_by_details(batch_details, models.BatchDetail, "batch_id", enums.EntityType.BATCH, None)

        batch_matches = self.db.query(models.Batch).filter(models.Batch.id.in_(select(subq))).all()
        return self._check_single_result(batch_matches, "batches")

    def _lookup_assay_run_by_details(
        self, assay_filter: Dict[str, Any], assay_run_details: Dict[str, Any]
    ) -> models.AssayRun:
        assay_query = self.db.query(models.Assay)
        for col_name, val in assay_filter.items():
            col = getattr(models.Assay, col_name, None)
            if col is None:
                continue
            assay_query = assay_query.filter(col == val)

        assays = assay_query.all()
        assay = self._check_single_result(assays, "assays")
        subq = self._lookup_by_details(
            assay_run_details, models.AssayRunDetail, "assay_run_id", enums.EntityType.ASSAY_RUN, None
        )
        assay_runs = (
            self.db.query(models.AssayRun)
            .filter(models.AssayRun.assay_id == assay.id)
            .filter(models.AssayRun.id.in_(select(subq)))
            .all()
        )
        return self._check_single_result(assay_runs, "assay runs")

    def _build_assay_result_record(self, batch_id: int, assay_run_id: int) -> Dict[str, Any]:
        return {
            "batch_id": batch_id,
            "assay_run_id": assay_run_id,
            "created_by": admin.admin_user_id,
            "updated_by": admin.admin_user_id,
        }

    # TODO: Identify the specific data row(s) in assay_results.csv causing failures
    def build_sql(self, rows: List[Dict[str, Any]]) -> str:
        self.assay_results_to_insert = []
        details = []

        for idx, row in enumerate(rows):
            try:
                grouped = self._group_data(row, "assay")
                batch_record = self._lookup_batch_by_details(grouped.get("batch_details"))
                assay_run_record = self._lookup_assay_run_by_details(
                    grouped.get("assay"), grouped.get("assay_run_details")
                )

                batch_id = getattr(batch_record, "id")
                assay_run_id = getattr(assay_run_record, "id")
                assay_result = self._build_assay_result_record(batch_id, assay_run_id)
                self.assay_results_to_insert.append(assay_result)

                inserted, updated = self.property_service.build_details_records(
                    models.AssayResultDetail,
                    grouped.get("assay_result_details", {}),
                    {"rn": idx + 1},
                    enums.EntityType.ASSAY_RESULT,
                    False,
                )
                details.extend(inserted)
                self._add_output_row(row, "success")
            except Exception as e:
                self.handle_row_error(row, e, idx, rows)

        if self.assay_results_to_insert:
            batch_sql = self.generate_sql(self.assay_results_to_insert, details)
            details.clear()
            return batch_sql

    def generate_sql(self, assay_results, details) -> str:
        assay_results_sql = self._generate_assay_result_sql(assay_results)
        details_sql = self._generate_details_sql(details)
        return sql_utils.generate_sql(assay_results_sql, details_sql)

    def _generate_assay_result_sql(self, assay_results) -> str:
        cols = list(assay_results[0].keys())
        values_sql = sql_utils.values_sql(assay_results, cols)
        return f"""
            WITH inserted_assay_results AS (
                INSERT INTO moltrack.assay_results ({", ".join(cols)})
                VALUES {values_sql}
                RETURNING id
            ),
            numbered_assay_results AS (
                SELECT id, ROW_NUMBER() OVER (ORDER BY id) AS rn
                FROM inserted_assay_results
            )"""

    def _generate_details_sql(self, details) -> str:
        if not details:
            return ""

        cols_without_key, values_sql = sql_utils.prepare_sql_parts(details)
        return f"""
            inserted_assay_result_details AS (
                INSERT INTO moltrack.assay_result_details (assay_result_id, {", ".join(cols_without_key)})
                SELECT nr.id, {", ".join([f"d.{col}" for col in cols_without_key])}
                FROM (VALUES {values_sql}) AS d(rn, {", ".join(cols_without_key)})
                JOIN numbered_assay_results nr ON d.rn = nr.rn
            )"""

    def cleanup_chunk(self):
        super().cleanup_chunk()
        self.assay_results_to_insert.clear()

    def cleanup(self):
        super().cleanup()
        self.cleanup_chunk()



from app import models
from app.setup.database import SessionLocal


class AdminUser:
    def __init__(self):
        self._admin_user_id = None

    @property
    def admin_user_id(self):
        if self._admin_user_id is None:
            db = SessionLocal()
            try:
                admin = db.query(models.User).filter(models.User.email == "admin@datagrok.ai").first()
                if not admin:
                    raise Exception("Admin user not found.")
                self._admin_user_id = admin.id
            finally:
                db.close()
        return self._admin_user_id


admin = AdminUser()



from typing import List, Dict, Any
import re
from sqlalchemy.orm import Session
from app.models import Token
from app.utils.enums import CompareOp, LogicOp
from app.services.search.parser import Parser
from app.services.search.engine import SearchEngine


class SearchFilterBuilder:
    def __init__(self, db: Session):
        self.COMPARE_OPS = [op.value for op in CompareOp]
        self.COMPARE_OPS = sorted(self.COMPARE_OPS, key=lambda s: len(s.split()), reverse=True)
        self.LOGICAL_OPS = [op.value for op in LogicOp]
        self.db = db

    def build_filter(self, expression: str) -> Dict[str, Any]:
        expression = expression.replace('"', "'")
        tokens = self._tokenize(expression=expression)

        parser = Parser(tokens, self.LOGICAL_OPS, self.COMPARE_OPS)
        filter_tree = parser.parse()

        search_engine = SearchEngine(self.db)
        err = search_engine._validate_filter(filter_tree, "")
        return {"filter": filter_tree.model_dump() if err == [] else None, "errors": err}

    def _tokenize(self, expression: str) -> List[Token]:
        tokens: List[Token] = []
        i = 0
        length = len(expression)

        while i < length:
            c = expression[i]

            # Skip whitespace
            if c.isspace():
                i += 1
                continue

            # Parentheses
            if c in "()":
                tokens.append(Token("LPAREN" if c == "(" else "RPAREN", c))
                i += 1
                continue

            # Array literal
            if c == "[":
                end = i
                bracket_level = 1
                while end + 1 < length and bracket_level > 0:
                    end += 1
                    if expression[end] == "[":
                        bracket_level += 1
                    elif expression[end] == "]":
                        bracket_level -= 1
                if bracket_level != 0:
                    raise ValueError("Unclosed array literal")

                array_content = expression[i + 1 : end].strip()
                if not array_content:
                    values = []
                else:
                    raw_items = re.split(r",\s*", array_content)
                    values = []
                    for item in raw_items:
                        item = item.strip()
                        if item.startswith("'") and item.endswith("'"):
                            values.append(item[1:-1])
                        elif re.match(r"^-?\d+\.\d+$", item):
                            values.append(float(item))
                        elif re.match(r"^-?\d+$", item):
                            values.append(int(item))
                        elif item.lower() == "true":
                            values.append(True)
                        elif item.lower() == "false":
                            values.append(False)
                        else:
                            raise ValueError(f"Unsupported array item: {item}")

                tokens.append(Token("ARRAY_LITERAL", values))
                i = end + 1
                continue

            # String literal (single quotes)
            if c == "'":
                end = i + 1
                end = expression.find("'", i + 1)
                if end == -1:
                    raise ValueError("Unclosed string literal")
                tokens.append(Token("STRING_LITERAL", expression[i + 1 : end]))
                i = end + 1
                continue

            # Number literal
            number_match = re.match(r"-?\d+(\.\d+)?", expression[i:])
            if number_match:
                num_str = number_match.group(0)
                value = float(num_str) if "." in num_str else int(num_str)
                tokens.append(Token("NUMBER_LITERAL", value))
                i += len(num_str)
                continue

            # Boolean literal
            match = re.match(r"(true|false)", expression[i:], re.IGNORECASE)
            if match:
                tokens.append(Token("BOOLEAN_LITERAL", match.group(1).lower() == "true"))
                i += len(match.group(0))
                continue

            # Multi-word operators
            matched_op = next((op for op in self.COMPARE_OPS + self.LOGICAL_OPS if expression[i:].startswith(op)), None)
            if matched_op:
                token_type = "COMPARE_OP" if matched_op in self.COMPARE_OPS else "LOGICAL_OP"
                tokens.append(Token(token_type, matched_op))
                i += len(matched_op)
                continue

            # Field name (dot notation, alphanumeric + underscores)
            field_match = re.match(r"[a-zA-Z_][a-zA-Z0-9_.]*", expression[i:])
            if field_match:
                field = field_match.group(0)
                tokens.append(Token("FIELD", field))
                i += len(field)
                continue

            raise ValueError(f"Unexpected character at position {i}: '{c}'")
        return tokens



from tests.conftest import _preload_assays, _preload_assay_runs, BLACK_DIR


def test_create_assay(client, preload_schema):
    response = _preload_assays(client, BLACK_DIR / "assays.json")
    assert response.status_code == 200

    assays = response.json()["created"]
    assert isinstance(assays, list)
    assert len(assays) == 1
    assert assays[0]["name"] == "Hepatocyte Stability"


def test_get_all_assays(client, preload_schema, preload_assays):
    response = client.get("/v1/assays/")
    assert response.status_code == 200

    assays = response.json()
    assert isinstance(assays, list)
    assert len(assays) == 1

    assay = assays[0]
    assert assay["name"] == "Hepatocyte Stability"
    assert "properties" in assay and isinstance(assay["properties"], list) and len(assay["properties"]) > 0
    assert "assay_details" in assay and isinstance(assay["assay_details"], list) and len(assay["assay_details"]) > 0
    assert (
        "property_requirements" in assay
        and isinstance(assay["property_requirements"], list)
        and len(assay["property_requirements"]) > 0
    )


def test_create_assay_run(client, preload_schema, preload_assays):
    response = _preload_assay_runs(client, BLACK_DIR / "assay_runs.csv", BLACK_DIR / "assay_runs_mapping.json")
    assert response.status_code == 200
    result = response.json()

    assert isinstance(result, list)
    assert len(result) == 9
    assert all(run.get("registration_status") == "success" for run in result)


def test_get_assay_run_by_id(client, preload_schema, preload_assays, preload_assay_runs):
    response = client.get("/v1/assay_runs/2")
    assert response.status_code == 200

    data = response.json()
    assert data["id"] == 2
    assert data["name"] == "Hepatocyte Stability2024-02-02"
    assert data["assay_id"] == 1

    assay = data.get("assay")
    assert assay and assay["id"] == 1
    assert assay["name"] == "Hepatocyte Stability"

    details = data.get("assay_run_details")
    assert isinstance(details, list)
    assert len(details) == 3

    detail_12 = next((d for d in details if d["property_id"] == 12), None)
    assert detail_12 is not None
    assert detail_12["value_string"] == "Human"

    props = data.get("properties")
    assert isinstance(props, list)
    assert any(p["name"] == "Cell Species" for p in props)


def test_get_assay_runs(client, preload_schema, preload_assays, preload_assay_runs):
    response = client.get("/v1/assay_runs/")
    assert response.status_code == 200

    data = response.json()
    assert len(data) == 9

    first = data[0]
    assert first["name"] == "Hepatocyte Stability2024-02-01"
    assert first["assay_id"] == 1
    assert first["id"] == 1

    assay = first.get("assay")
    assert assay is not None
    assert assay.get("name") == "Hepatocyte Stability"

    details = first.get("assay_run_details", [])
    assert isinstance(details, list)
    assert len(details) > 0

    properties = first.get("properties", [])
    assert isinstance(properties, list)
    assert len(properties) > 0



from datetime import datetime
from typing import Any, Callable, Dict
import uuid
from dateutil.parser import parse


def cast_datetime(value: Any) -> datetime:
    if isinstance(value, datetime):
        return value
    try:
        return datetime.fromisoformat(value)
    except Exception:
        try:
            # Fall back to dateutil for flexible parsing
            return parse(value)
        except Exception:
            raise ValueError(f"Invalid datetime format: {value}")


def cast_bool(value: Any) -> bool:
    if isinstance(value, bool):
        return value
    if isinstance(value, str):
        val = value.strip().lower()
        if val in ("true", "1"):
            return True
        elif val in ("false", "0"):
            return False
    if isinstance(value, (int, float)):
        return bool(value)
    raise ValueError(f"Invalid boolean format: {value}")


def cast_uuid(value: Any) -> uuid.UUID:
    if isinstance(value, uuid.UUID):
        return value
    try:
        return uuid.UUID(str(value))
    except Exception:
        raise ValueError(f"Invalid UUID format: {value}")


value_type_to_field: Dict[str, str] = {
    "datetime": "value_datetime",
    "int": "value_num",
    "double": "value_num",
    "string": "value_string",
    "uuid": "value_uuid",
    "bool": "value_bool",
}

value_type_cast_map: Dict[str, Callable[[Any], Any]] = {
    "datetime": cast_datetime,
    "int": int,
    "double": float,
    "string": str,
    "uuid": cast_uuid,
    "bool": cast_bool,
}



import os
from sqlalchemy import create_engine
from sqlalchemy.orm import declarative_base
from sqlalchemy.orm import sessionmaker

# Database connection parameters
DB_USER = os.environ.get("DB_USER", "postgres")
DB_PASSWORD = os.environ.get("DB_PASSWORD", "postgres")
DB_HOST = os.environ.get("DB_HOST", "localhost")
DB_PORT = os.environ.get("DB_PORT", "5432")
DB_NAME = os.environ.get("DB_NAME", "moltrack")
DB_SCHEMA = os.environ.get("DB_SCHEMA", "moltrack")

# Construct the database URL from the parameters
SQLALCHEMY_DATABASE_URL = f"postgresql://{DB_USER}:{DB_PASSWORD}@{DB_HOST}:{DB_PORT}/{DB_NAME}"

# Create the engine with the appropriate URL
engine = create_engine(SQLALCHEMY_DATABASE_URL, connect_args={"options": f"-csearch_path={DB_SCHEMA},public"})
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)

Base = declarative_base()



"""
Search operators and their SQL translations
"""

from typing import Dict, Any, Tuple
from datetime import datetime
from app.utils.enums import OperatorType


class SearchOperators:
    """Maps search operators to SQL expressions and validation"""

    # TODO: CHECK OPERATORS FOR VALIDITY AND CONSISTENCY
    # Operator definitions with their SQL translations and parameter handling
    OPERATORS = {
        # String operators
        "=": {"sql": "= :param", "type": OperatorType.STRING, "params": 1, "description": "Exact match"},
        "!=": {"sql": "!= :param", "type": OperatorType.STRING, "params": 1, "description": "Not equal"},
        "LIKE": {
            "sql": "LIKE :param",
            "type": OperatorType.STRING,
            "params": 1,
            "description": "Pattern match with wildcards",
        },
        "CONTAINS": {
            "sql": "ILIKE :param",
            "type": OperatorType.STRING,
            "params": 1,
            "description": "Case-insensitive contains",
            "value_transform": lambda x: f"%{x}%",
        },
        "STARTS WITH": {
            "sql": "ILIKE :param",
            "type": OperatorType.STRING,
            "params": 1,
            "description": "Starts with pattern",
            "value_transform": lambda x: f"{x}%",
        },
        "ENDS WITH": {
            "sql": "ILIKE :param",
            "type": OperatorType.STRING,
            "params": 1,
            "description": "Ends with pattern",
            "value_transform": lambda x: f"%{x}",
        },
        "IN": {
            "sql": "{field} IN {params}",
            "type": OperatorType.STRING,
            "params": 1,
            "description": "Match any value in list",
            "value_transform": lambda x: tuple(x) if isinstance(x, list) else x,
        },
        # Numeric operators
        "<": {"sql": "< :param", "type": OperatorType.NUMERIC, "params": 1, "description": "Less than"},
        ">": {"sql": "> :param", "type": OperatorType.NUMERIC, "params": 1, "description": "Greater than"},
        "<=": {"sql": "<= :param", "type": OperatorType.NUMERIC, "params": 1, "description": "Less than or equal"},
        ">=": {"sql": ">= :param", "type": OperatorType.NUMERIC, "params": 1, "description": "Greater than or equal"},
        "RANGE": {
            "sql": "{field} BETWEEN :param1 AND :param2",
            "type": OperatorType.NUMERIC,
            "params": 2,
            "description": "Between two values (inclusive)",
            "value_transform": lambda x: (x[0], x[1]) if isinstance(x, (list, tuple)) and len(x) == 2 else x,
        },
        # Datetime operators
        "BEFORE": {"sql": "< :param", "type": OperatorType.DATETIME, "params": 1, "description": "Before date/time"},
        "AFTER": {"sql": "> :param", "type": OperatorType.DATETIME, "params": 1, "description": "After date/time"},
        "ON": {
            "sql": "DATE({field}) = DATE(:param)",
            "type": OperatorType.DATETIME,
            "params": 1,
            "description": "On specific date",
        },
        # Molecular operators (RDKit)
        "IS SIMILAR": {
            "sql": "public.tanimoto_sml(public.morganbv_fp(public.mol_from_smiles({field}::cstring)), "
            "public.morganbv_fp(public.mol_from_smiles(:param2))) >= :param3",
            "type": OperatorType.MOLECULAR,
            "params": 1,
            "description": "Molecular similarity using RDKit",
            "requires_threshold": True,
        },
        "HAS SUBSTRUCTURE": {
            "sql": "public.mol_from_smiles({field}::cstring) OPERATOR(public.@>) :param\\:\\:public.qmol",
            "type": OperatorType.MOLECULAR,
            "params": 1,
            "description": "Molecular substructure search",
        },
        "IS SUBSTRUCTURE OF": {
            "sql": "public.mol_from_smiles(:param) OPERATOR(public.@>) public.mol_from_smiles({field}::cstring)",
            "type": OperatorType.MOLECULAR,
            "params": 1,
            "description": "Molecular superstructure search",
        },
    }

    @classmethod
    def get_operator(cls, operator: str) -> Dict[str, Any]:
        """Get operator definition"""
        if operator not in cls.OPERATORS:
            raise ValueError(f"Unsupported operator: {operator}")
        return cls.OPERATORS[operator]

    @classmethod
    def validate_operator_value(cls, operator: str, value: Any, threshold: float = None) -> bool:
        """Validate that a value is appropriate for the given operator"""
        op_def = cls.get_operator(operator)

        # Check if threshold is required
        if op_def.get("requires_threshold", False) and threshold is None:
            raise ValueError(f"Operator '{operator}' requires a threshold value")

        # Check value type based on operator
        if operator == "IN" and not isinstance(value, (list, tuple)):
            raise ValueError(f"Operator '{operator}' requires a list or tuple value")

        if operator == "RANGE":
            if not isinstance(value, (list, tuple)) or len(value) != 2:
                raise ValueError(f"Operator '{operator}' requires a list/tuple with exactly 2 values")
            if value[0] >= value[1]:
                raise ValueError("RANGE operator requires first value to be less than second value")

        if operator in ["ON", "BEFORE", "AFTER"]:

            def check_date(date_str: str):
                formats = ["%Y-%m-%d", "%Y-%m-%d %H:%M"]  # Supported formats: date-only and date-time
                for fmt in formats:
                    try:
                        # Try to parse the date string using the YYYY-MM-DD format
                        datetime.strptime(date_str, fmt)
                        return True
                    except ValueError:
                        continue
                return False

            if not check_date(value):
                raise ValueError("DATE value must be in format YYYY-MM-DD or YYYY-MM-DD hh:mm")
        return True

    @classmethod
    def validate_operands(cls, operator: str, field: str):
        """Validate that operand is appropriate for operation"""
        if operator in ["IS SIMILAR", "HAS SUBSTRUCTURE", "IS SUBSTRUCTURE OF"]:
            if not field.endswith(".structure"):
                raise ValueError("Molecular operators can only be applied to compounds.structure")
        else:
            if field.endswith(".structure"):
                raise ValueError("Only molecular operators can be applied to compounds.structure")

    @classmethod
    def get_sql_expression(
        cls, operator: str, field: str, value: Any, threshold: float = None
    ) -> Tuple[str, Dict[str, Any]]:
        """
        Get SQL expression and parameters for an operator

        Returns:
            Tuple of (sql_expression, parameters_dict)
        """
        op_def = cls.get_operator(operator)

        # Transform value if needed
        if "value_transform" in op_def:
            value = op_def["value_transform"](value)

        # Handle special cases
        match operator:
            case "IN":
                # For IN clauses, we need to create multiple parameters
                param_names = [f"param{i + 1}" for i in range(len(value))]
                placeholders = "(" + ",".join([f":{name}" for name in param_names]) + ")"
                sql_expr = op_def["sql"].format(field=field, params=placeholders)
                params = {name: val for name, val in zip(param_names, value)}
                return sql_expr, params
            case "RANGE":
                sql_expr = op_def["sql"].format(field=field)
                return sql_expr, {"param1": value[0], "param2": value[1]}
            case "IS SIMILAR":
                sql_expr = op_def["sql"].format(field=field)
                return sql_expr, {"param1": value, "param2": value, "param3": threshold}
            case "IS SUBSTRUCTURE OF" | "HAS SUBSTRUCTURE" | "ON":
                sql_expr = op_def["sql"].format(field=field)
                return sql_expr, {"param": value}
            case _:
                # Standard operator
                sql_expr = f"{field} {op_def['sql']}"
                return sql_expr, {"param": value}



from typing import Any, Dict, List, Optional

from fastapi import HTTPException
from sqlalchemy.orm import Session

from app.utils.admin_utils import admin
from app import models
from app.utils import enums, sql_utils
from app.services.registrars.base_registrar import BaseRegistrar


class AssayRunRegistrar(BaseRegistrar):
    def __init__(self, db: Session, mapping: Optional[str], error_handling: str):
        super().__init__(db, mapping, error_handling)
        self._assay_records_map = None
        self.assay_runs_to_insert = []

    @property
    def assay_records_map(self):
        if self._assay_records_map is None:
            self._assay_records_map = self._load_reference_map(models.Assay, "name")
        return self._assay_records_map

    def _build_assay_run_record(self, assay_data: Dict[str, Any], assay_details: Dict[str, Any]) -> Dict[str, Any]:
        assay_name = assay_data.get("name")
        existing_assay = self.assay_records_map.get(assay_name)
        if not existing_assay:
            raise HTTPException(status_code=400, detail=f"Assay {assay_name} not found.")

        # TODO: Clarify and implement the rules for creating the 'name' attribute in assay_run entries
        return {
            "assay_id": getattr(existing_assay, "id"),
            "name": assay_name + assay_details.get("Assay Run Date"),
            "created_by": admin.admin_user_id,
            "updated_by": admin.admin_user_id,
        }

    def build_sql(self, rows: List[Dict[str, Any]]) -> str:
        self.assay_runs_to_insert = []
        details = []

        for idx, row in enumerate(rows):
            try:
                grouped = self._group_data(row, "assay")
                assay_data = grouped.get("assay", {})
                assay_run = self._build_assay_run_record(assay_data, grouped.get("assay_run_details"))
                self.assay_runs_to_insert.append(assay_run)

                inserted, updated = self.property_service.build_details_records(
                    models.AssayRunDetail,
                    grouped.get("assay_run_details", {}),
                    {"rn": idx + 1},
                    enums.EntityType.ASSAY_RUN,
                    False,
                )
                details.extend(inserted)
                self._add_output_row(row, "success")
            except Exception as e:
                self.handle_row_error(row, e, idx, rows)

        if self.assay_runs_to_insert:
            batch_sql = self.generate_sql(self.assay_runs_to_insert, details)
            details.clear()
            return batch_sql

    def generate_sql(self, assay_runs, details) -> str:
        assay_runs_sql = self._generate_assay_run_sql(assay_runs)
        details_sql = self._generate_details_sql(details)
        return sql_utils.generate_sql(assay_runs_sql, details_sql)

    # TODO: Think of a more robust key than row number for joining
    def _generate_assay_run_sql(self, assay_runs) -> str:
        cols = list(assay_runs[0].keys())
        values_sql = sql_utils.values_sql(assay_runs, cols)
        return f"""
            WITH inserted_assay_runs AS (
                INSERT INTO moltrack.assay_runs ({", ".join(cols)})
                VALUES {values_sql}
                RETURNING id
            ),
            numbered_assay_runs AS (
                SELECT id, ROW_NUMBER() OVER (ORDER BY id) AS rn
                FROM inserted_assay_runs
            )"""

    def _generate_details_sql(self, details) -> str:
        if not details:
            return ""

        cols_without_key, values_sql = sql_utils.prepare_sql_parts(details)
        return f"""
            inserted_assay_run_details AS (
                INSERT INTO moltrack.assay_run_details (assay_run_id, {", ".join(cols_without_key)})
                SELECT nr.id, {", ".join([f"d.{col}" for col in cols_without_key])}
                FROM (VALUES {values_sql}) AS d(rn, {", ".join(cols_without_key)})
                JOIN numbered_assay_runs nr ON d.rn = nr.rn
            )"""

    def cleanup_chunk(self):
        super().cleanup_chunk()
        self.assay_runs_to_insert.clear()

    def cleanup(self):
        super().cleanup()
        self.cleanup_chunk()
        self._assay_records_map = None



import csv
from datetime import datetime
from io import BytesIO, StringIO
import json
import re
from typing import Any, List
from fastapi import Response
import pandas as pd
from sqlalchemy.orm import Session
from sqlalchemy import text
from app.models import Level
from app.utils import enums


def create_alias(table: Level) -> str:
    if table != enums.SearchEntityType.ASSAY_RUNS.value:
        name_parts = table.split("_")
        if len(name_parts) == 2:
            # Since the table name is in the format "table_name", we can use the first letter of the first part and the
            # first letter of the second part
            return f"{name_parts[0][0]}{name_parts[1][0]}"
        else:
            return f"{name_parts[0][0]}"
    else:
        return "rn"


def singularize(word: str) -> str:
    if word.endswith("es"):
        return word[:-2]
    elif word.endswith("s"):
        return word[:-1]
    return word


def get_table_columns(table_name: str, session: Session) -> list:
    """
    Get the columns of a table in the database.
    """

    result = session.execute(
        text("SELECT column_name FROM information_schema.columns WHERE table_name = :table_name"),
        {"table_name": table_name},
    )
    columns = [row[0] for row in result.fetchall()]
    return columns


def sanitize_field_name(field_name: str) -> str:
    """Sanitize field name for use in SQL aliases"""
    # Replace dots and special characters with underscores
    sanitized = re.sub(r"[^a-zA-Z0-9_]", "_", field_name)
    # Ensure it doesn't start with a number
    if sanitized and sanitized[0].isdigit():
        sanitized = f"field_{sanitized}"
    return sanitized


def convert_datetime(obj):
    if isinstance(obj, datetime):
        return obj.isoformat()
    return obj


def prepare_search_output(results: List[Any], headers: List[str], output_format: enums.SearchOutputFormat):
    match output_format:
        case enums.SearchOutputFormat.json:
            return_obj = {
                "status": "success",
                "total_count": len(results),
                "columns": headers,
                "data": [dict(zip(headers, row)) for row in results],
            }
            json_output = json.dumps(return_obj, default=convert_datetime)
            return Response(
                content=json_output,
                media_type="application/json",
                headers={"Content-Disposition": "attachment; filename=result.json"},
            )
        case enums.SearchOutputFormat.csv:
            output = StringIO()
            writer = csv.writer(output)
            writer.writerow(headers)
            writer.writerows(results)
            csv_content = output.getvalue()
            return Response(
                content=csv_content,
                media_type="text/csv",
                headers={"Content-Disposition": "attachment; filename=result.csv"},
            )
        case enums.SearchOutputFormat.parquet:
            df = pd.DataFrame(results, columns=headers)
            buffer = BytesIO()
            df.to_parquet(buffer, engine="pyarrow", index=False)
            return Response(
                content=buffer.getvalue(),
                media_type="application/octet-stream",
                headers={"Content-Disposition": "attachment; filename=result.parquet"},
            )



# app/crud/__init__.py

from .compounds import (
    get_compound_by_id,
    get_compound_by_hash,
    read_compounds,
    update_compound,
    delete_compound,
)

from .properties import create_properties, get_properties, get_entities_by_entity_type, get_synonym_id

from .batches import get_batches_by_compound, get_batches, get_batch, delete_batch

from .additions import (
    create_additions,
    get_additions,
    get_addition_by_id,
    update_addition_by_id,
    delete_addition_by_id,
    get_batch_addition_for_addition,
)

from .assay_data import get_assays, get_assay, get_assay_runs, get_assay_run, get_all_assay_results_for_batch



import csv
import io
import json
from abc import ABC, abstractmethod
from typing import Iterator, List, Dict, Any, Optional

from sqlalchemy import select, text
from fastapi import HTTPException

from app.utils import enums
from app.utils.logging_utils import logger
from app.services import property_service
from app import models


class BaseRegistrar(ABC):
    def __init__(self, db, mapping: Optional[str], error_handling: str):
        """
        Base class for processing and registering data to a database.
        :param db: SQLAlchemy database session.
        :param mapping: Optional JSON string defining field mappings.
        :param error_handling: Strategy for handling errors during processing.
        """
        self.db = db
        self.error_handling = error_handling
        self._property_records_map = None
        self._addition_records_map = None

        self.property_service = property_service.PropertyService(self.property_records_map)
        self.user_mapping = self._load_mapping(mapping)
        self.output_rows = []

    @property
    def property_records_map(self):
        if self._property_records_map is None:
            self._property_records_map = self._load_reference_map(models.Property, "name")
        return self._property_records_map

    @property
    def addition_records_map(self):
        if self._addition_records_map is None:
            self._addition_records_map = self._load_reference_map(models.Addition, "name")
        return self._addition_records_map

    # === Input processing methods ===

    def _load_mapping(self, mapping: Optional[str]) -> Dict[str, str]:
        if not mapping:
            return {}
        try:
            return json.loads(mapping)
        except json.JSONDecodeError:
            raise HTTPException(status_code=400, detail="Invalid JSON for mapping")

    def process_csv(self, file_stream: io.TextIOBase, chunk_size=5000) -> Iterator[List[Dict[str, Any]]]:
        reader = csv.DictReader(file_stream, skipinitialspace=True)

        try:
            first_row = next(reader)
        except StopIteration:
            raise HTTPException(status_code=400, detail="CSV file is empty or invalid")

        if self.user_mapping:
            self.normalized_mapping = self.user_mapping
        else:
            self.normalized_mapping = {}
            for col in first_row.keys():
                assigned = self._assign_column(col)
                self.normalized_mapping[col] = assigned

        chunk = [first_row]

        for row in reader:
            chunk.append(row)
            if len(chunk) >= chunk_size:
                yield chunk
                chunk = []

        if chunk:
            yield chunk

    def _assign_column(self, col: str) -> str:
        if col in self.property_records_map:
            entity_type = getattr(self.property_records_map[col], "entity_type", None)
            prefix = {
                enums.EntityType.COMPOUND: "compound_details",
                enums.EntityType.BATCH: "batch_details",
                enums.EntityType.ASSAY_RUN: "assay_run_details",
                enums.EntityType.ASSAY_RESULT: "assay_result_details",
            }.get(entity_type)
            return f"{prefix}.{col}" if prefix else col

        if col in self.addition_records_map:
            return f"batch_additions.{col}"
        return col

    def _group_data(self, row: Dict[str, Any], entity_name: Optional[str] = None) -> Dict[str, Dict[str, Any]]:
        grouped = {}
        for src_key, mapped_key in self.normalized_mapping.items():
            value = row.get(src_key)
            table, field = (
                mapped_key.split(".", 1)
                if "." in mapped_key
                else (entity_name if entity_name else "compound", mapped_key)
            )
            grouped.setdefault(table, {})[field] = value
        return grouped

    # === Reference loading methods ===

    def _load_reference_map(self, model, key: str = "id"):
        result = self.db.execute(select(model)).scalars().all()
        return {getattr(row, key): row for row in result}

    def model_to_dict(self, obj):
        return {c.key: getattr(obj, c.key) for c in obj.__table__.columns}

    # === SQL construction and registration methods ===

    @abstractmethod
    def build_sql(self, rows: List[Dict[str, Any]]) -> str:
        pass

    @abstractmethod
    def generate_sql(self) -> Optional[str]:
        pass

    def register_all(self, rows: List[Dict[str, Any]]):
        batch_sql = self.build_sql(rows)

        if batch_sql:
            try:
                self.db.execute(text(batch_sql))
                self.db.commit()
            except Exception as e:
                logger.error(f"An exception occurred: {e}")
                self.db.rollback()

    # === Output formatting methods ===

    def _add_output_row(self, row, status, error_msg=None):
        row["registration_status"] = status
        row["registration_error_message"] = error_msg or ""
        self.output_rows.append(row)

    def cleanup_chunk(self):
        self.output_rows.clear()

    def cleanup(self):
        self.cleanup_chunk()
        self.user_mapping.clear()
        self.normalized_mapping.clear()
        self._property_records_map = None
        self._addition_records_map = None
        self.property_service = None

    # === Error handling methods ===

    def handle_row_error(self, row, exception, global_idx, all_rows):
        self._add_output_row(row, "failed", str(exception))
        if self.error_handling == enums.ErrorHandlingOptions.reject_all.value:
            for remaining_row in all_rows[global_idx + 1 :]:
                self._add_output_row(remaining_row, "not_processed")
            raise HTTPException(status_code=400, detail="Registration failed. See the output file for details.")



from typing import List
from sqlalchemy.orm import Session
from app import models


# === Assay-related operations ===
def get_assay(db: Session, assay_id: int):
    return db.query(models.Assay).filter(models.Assay.id == assay_id).first()


def get_assays(db: Session, skip: int = 0, limit: int = 100):
    return db.query(models.Assay).offset(skip).limit(limit).all()


# === AssayRun-related operations ===
def get_assay_run(db: Session, assay_run_id: int):
    # Get the assay
    assay = db.query(models.AssayRun).filter(models.AssayRun.id == assay_run_id).first()

    if assay:
        # Get properties associated with this assay through assay details
        assay_details = db.query(models.AssayRunDetail).filter(models.AssayRunDetail.assay_run_id == assay_run_id).all()
        property_ids = [detail.property_id for detail in assay_details]

        # If no properties from assay details, get them from the assay type
        if not property_ids and assay.assay:
            assay_properties = (
                db.query(models.AssayProperty).filter(models.AssayProperty.assay_id == assay.assay_id).all()
            )
            property_ids = [prop.property_id for prop in assay_properties]

        # Get the property objects
        if property_ids:
            properties = db.query(models.Property).filter(models.Property.id.in_(property_ids)).all()
            # Add properties to assay
            assay.properties = properties
        else:
            assay.properties = []

    return assay


def get_assay_runs(db: Session, skip: int = 0, limit: int = 100):
    # Get assay_runs with pagination
    assay_runs = db.query(models.AssayRun).offset(skip).limit(limit).all()

    # For each assay, add its properties
    for assay_run in assay_runs:
        # Get properties associated with this assay through assay details
        assay_details = db.query(models.AssayRunDetail).filter(models.AssayRunDetail.assay_run_id == assay_run.id).all()
        property_ids = [detail.property_id for detail in assay_details]

        # If no properties from assay details, get them from the assay type
        if not property_ids and assay_run.assay:
            assay_properties = (
                db.query(models.AssayProperty).filter(models.AssayProperty.assay_id == assay_run.assay_id).all()
            )
            property_ids = [prop.property_id for prop in assay_properties]

        # Get the property objects
        if property_ids:
            properties = db.query(models.Property).filter(models.Property.id.in_(property_ids)).all()
            # Add properties to assay
            assay_run.properties = properties
        else:
            assay_run.properties = []

    return assay_runs


# === AssayResult-related operations ===
def get_assay_result(db: Session, assay_result_id: int):
    return db.query(models.AssayResult).filter(models.AssayResult.id == assay_result_id).first()


def get_assay_results(db: Session, skip: int = 0, limit: int = 100):
    return db.query(models.AssayResult).offset(skip).limit(limit).all()


def get_all_assay_results_for_batch(db: Session, batch_id: int) -> List[models.AssayResult]:
    return db.query(models.AssayResult).filter(models.AssayResult.batch_id == batch_id).all()


def get_batch_assay_results(db: Session, batch_id: int):
    """Get all assay results for a specific batch"""
    results = get_all_assay_results_for_batch(db, batch_id)

    # Group results by assay_run_id
    grouped_results = {}
    for result in results:
        assay_run_id = result.assay_run_id
        if assay_run_id not in grouped_results:
            # Get the assay name
            assay_run = db.query(models.AssayRun).filter(models.AssayRun.id == assay_run_id).first()
            assay = db.query(models.Assay).filter(models.Assay.id == assay_run.assay_id).first()
            assay_name = assay.name if assay else "Unknown Assay"

            grouped_results[assay_run_id] = {
                "assay_run_id": assay_run_id,
                "batch_id": batch_id,
                "assay_name": assay_name,
                "measurements": {},
            }

        # Get property name and type
        property = db.query(models.Property).filter(models.Property.id == result.property_id).first()
        property_name = property.name if property else f"Property-{result.property_id}"
        property_type = property.value_type if property else "double"

        # Get value based on property type
        value = None
        if property_type in ("int", "double"):
            value = result.value_num
        elif property_type == "string":
            value = result.value_string
        elif property_type == "bool":
            value = result.value_bool

        # If we have a qualifier other than "=" (0), include it in the result
        if result.value_qualifier != 0:
            grouped_results[assay_run_id]["measurements"][property_name] = {
                "qualifier": result.value_qualifier,
                "value": value,
            }
        else:
            grouped_results[assay_run_id]["measurements"][property_name] = value

    return list(grouped_results.values())



from typing import List, Union
from pydantic import ValidationError
from app.models import AtomicCondition, LogicalNode, Token, Filter
from app.utils.enums import CompareOp, LogicOp


class ParseError(Exception):
    """Raised on invalid syntax or validation issues during parsing."""

    pass


class Parser:
    def __init__(self, tokens: List[Token], LOGICAL_OPS: List[str], COMPARE_OPS: List[str]):
        self.tokens = tokens
        self.pos = 0
        self.ops = {
            "LOGICAL_OP": LOGICAL_OPS,
            "COMPARE_OP": COMPARE_OPS,
        }

    def peek(self) -> Union[Token, None]:
        return self.tokens[self.pos] if self.pos < len(self.tokens) else None

    def advance(self) -> Token:
        tok = self.peek()
        self.pos += 1
        return tok

    def expect(self, expected_type: str) -> Token:
        tok = self.peek()
        if not tok or tok.type != expected_type:
            raise ParseError(
                f"Expected token of type '{expected_type}', but got {tok}. Possible values: {self.ops[expected_type]}"
            )
        return self.advance()

    def parse(self) -> Filter:
        result = self.parse_expression()
        if self.peek() is not None:
            raise ParseError(f"Unexpected trailing token: {self.peek()}")
        return result

    def parse_expression(self) -> Filter:
        left = self.parse_term()

        while self.peek() and self.peek().type == "LOGICAL_OP":
            op_token = self.advance()
            logic_op_str = op_token.value

            try:
                logic_op = LogicOp(logic_op_str)
            except ValueError:
                raise ParseError(f"Invalid logical operator: {logic_op_str}")

            right = self.parse_term()

            if isinstance(left, LogicalNode) and left.operator == logic_op:
                left.conditions.append(right)
            else:
                left = LogicalNode(operator=logic_op, conditions=[left, right])

        return left

    def parse_term(self) -> Filter:
        tok = self.peek()

        if tok and tok.type == "LPAREN":
            self.advance()
            node = self.parse_expression()
            if not self.peek() or self.peek().type != "RPAREN":
                raise ParseError("Expected closing parenthesis ')'")
            self.advance()
            return node
        else:
            return self.parse_atomic()

    def parse_atomic(self) -> AtomicCondition:
        field_token = self.expect("FIELD")
        operator_token = self.expect("COMPARE_OP")
        value_token = self.advance()

        if not value_token or value_token.type not in {
            "STRING_LITERAL",
            "NUMBER_LITERAL",
            "BOOLEAN_LITERAL",
            "ARRAY_LITERAL",
        }:
            raise ParseError(f"Expected a value literal, got {value_token}")

        field = field_token.value
        operator_str = operator_token.value

        try:
            operator = CompareOp(operator_str)
        except ValueError:
            raise ParseError(f"Unsupported comparison operator: {operator_str}")

        value = value_token.value
        threshold = None

        # Optional threshold for IS SIMILAR
        if operator == CompareOp.IS_SIMILAR:
            if self.peek() and self.peek().type == "NUMBER_LITERAL":
                threshold_token = self.advance()
                threshold = float(threshold_token.value)
            else:
                raise ParseError("IS SIMILAR operator requires a numeric threshold after the value")

        try:
            return AtomicCondition(field=field, operator=operator, value=value, threshold=threshold)
        except ValidationError as e:
            raise ParseError(f"Validation failed: {e}")



from typing import Any, Dict, List, NamedTuple, Optional, Union, Literal, get_args
from pydantic import ConfigDict, field_validator, model_validator
from sqlalchemy import Column, DateTime, Enum, CheckConstraint
from sqlmodel import SQLModel, Field, Relationship
from sqlalchemy.sql import func
from app.utils import enums
import os
from datetime import datetime
import uuid
# import crud
# # Handle both package imports and direct execution
# try:
#     # When imported as a package (for tests)
#     from .database import Base
# except ImportError:
#     # When run directly
#     from database import Base

# Get the schema name from environment variable or use default
DB_SCHEMA = os.environ.get("DB_SCHEMA", "moltrack")


class User(SQLModel, table=True):
    __tablename__ = "users"
    __table_args__ = {"schema": DB_SCHEMA}

    id: uuid.UUID = Field(primary_key=True, nullable=False, default_factory=uuid.uuid4)
    email: str = Field(nullable=False)
    first_name: str = Field(nullable=False)
    last_name: str = Field(nullable=False)
    has_password: bool = Field(nullable=False)
    is_active: bool = Field(default=False, nullable=False)
    is_service_account: bool = Field(default=False, nullable=False)
    created_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now(), nullable=False))
    updated_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now(), onupdate=func.now(), nullable=False)
    )
    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)


class Settings(SQLModel, table=True):
    __tablename__ = "settings"
    __table_args__ = {"schema": DB_SCHEMA}

    id: int = Field(primary_key=True, nullable=False)
    name: str = Field(nullable=False)
    value: str = Field(nullable=False)
    description: str = Field(nullable=False)


class CompoundQueryParams(SQLModel):
    substructure: Optional[str] = None
    skip: int = 0
    limit: int = 100


class CompoundBase(SQLModel):
    original_molfile: Optional[str] = Field(default=None)  # as sketched by the chemist
    is_archived: Optional[bool] = Field(default=False)


class CompoundCreate(CompoundBase):
    smiles: str


class CompoundResponseBase(CompoundBase):
    id: int = Field(primary_key=True, index=True)
    canonical_smiles: str = Field(nullable=False)
    inchi: str = Field(nullable=False)
    inchikey: str = Field(nullable=False, unique=True)
    created_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now()))
    updated_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now(), onupdate=func.now())
    )


class CompoundResponse(CompoundResponseBase):
    batches: List["Batch"] = []
    properties: Optional[List["PropertyWithValue"]] = []


class CompoundDetailBase(SQLModel):
    compound_id: int = Field(foreign_key="moltrack.compounds.id", nullable=False)
    property_id: int = Field(foreign_key="moltrack.properties.id", nullable=False)


class CompoundDetailCreate(CompoundDetailBase):
    value: Any


class CompoundDetail(CompoundDetailBase, table=True):
    __tablename__ = "compound_details"
    __table_args__ = {"schema": DB_SCHEMA}

    id: int = Field(primary_key=True, index=True)
    created_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now(), nullable=False))
    updated_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now(), onupdate=func.now(), nullable=False)
    )
    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)

    value_datetime: Optional[datetime] = Field(sa_column=Column(DateTime(timezone=True)))
    value_uuid: Optional[uuid.UUID] = Field(default_factory=uuid.uuid4)
    value_num: Optional[float]
    value_string: Optional[str]

    compound: "Compound" = Relationship(back_populates="compound_details")
    property: "Property" = Relationship(back_populates="compound_details")


class Compound(CompoundResponseBase, table=True):
    __tablename__ = "compounds"
    __table_args__ = {"schema": DB_SCHEMA}

    molregno: int = Field(nullable=False)
    formula: str = Field(nullable=False)
    hash_mol: str = Field(nullable=False)
    hash_tautomer: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    hash_canonical_smiles: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    hash_no_stereo_smiles: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    hash_no_stereo_tautomer: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    deleted_at: Optional[datetime] = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now()))
    deleted_by: Optional[uuid.UUID] = Field(default=None)

    batches: List["Batch"] = Relationship(back_populates="compound")
    compound_details: List["CompoundDetail"] = Relationship(back_populates="compound")
    properties: List["Property"] = Relationship(
        back_populates="compounds", link_model=CompoundDetail, sa_relationship_kwargs={"viewonly": True}
    )


class CompoundDetailUpdate(SQLModel):
    property_id: int
    value: Any


class CompoundUpdate(SQLModel):
    original_molfile: Optional[str] = None
    is_archived: Optional[bool] = None
    canonical_smiles: Optional[str] = None
    properties: Optional[List[CompoundDetailUpdate]] = None


class BatchDetailBase(SQLModel):
    batch_id: int = Field(foreign_key=f"{DB_SCHEMA}.batches.id", nullable=False)
    property_id: int = Field(foreign_key=f"{DB_SCHEMA}.properties.id", nullable=False)

    value_qualifier: Optional[int] = Field(default=0, nullable=False)  # 0 for "=", 1 for "<", 2 for ">"
    value_datetime: Optional[datetime] = Field(default=None, sa_column=Column(DateTime(timezone=True)))
    value_uuid: Optional[uuid.UUID] = Field(default_factory=uuid.uuid4)
    value_num: Optional[float] = Field(default=None)
    value_string: Optional[str] = Field(default=None)


class BatchDetailResponse(BatchDetailBase):
    id: int = Field(primary_key=True, index=True)


class BatchDetail(BatchDetailResponse, table=True):
    __tablename__ = "batch_details"
    __table_args__ = {"schema": DB_SCHEMA}

    created_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now(), nullable=False))
    updated_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now(), nullable=False))
    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)

    batch: "Batch" = Relationship(back_populates="batch_details")
    property: "Property" = Relationship(back_populates="batch_details")


class BatchBase(SQLModel):
    notes: Optional[str] = Field(default=None)
    compound_id: int = Field(foreign_key=f"{DB_SCHEMA}.compounds.id", nullable=False)


class BatchResponseBase(BatchBase):
    id: int = Field(primary_key=True, index=True)
    created_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now(), nullable=False))


class BatchResponse(BatchResponseBase):
    batch_additions: List["BatchAddition"] = []
    properties: Optional[List["PropertyWithValue"]] = []


class Batch(BatchResponseBase, table=True):
    __tablename__ = "batches"
    __table_args__ = {"schema": DB_SCHEMA}

    updated_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now(), nullable=False))
    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    batch_regno: int = Field(nullable=False)

    compound: "Compound" = Relationship(back_populates="batches")
    assay_results: List["AssayResult"] = Relationship(back_populates="batch")
    batch_details: List["BatchDetail"] = Relationship(back_populates="batch")
    batch_additions: List["BatchAddition"] = Relationship(back_populates="batch")
    properties: List["Property"] = Relationship(
        back_populates="batches", link_model=BatchDetail, sa_relationship_kwargs={"viewonly": True}
    )


class SemanticTypeBase(SQLModel):
    name: str = Field(nullable=False)
    description: Optional[str] = Field(default=None, nullable=True)


class SemanticType(SemanticTypeBase, table=True):
    __tablename__ = "semantic_types"
    __table_args__ = {"schema": DB_SCHEMA}

    id: int = Field(primary_key=True, nullable=False)


class PropertyBase(SQLModel):
    name: str = Field(nullable=False)
    description: Optional[str] = Field(default=None)
    value_type: enums.ValueType = Field(sa_column=Column(Enum(enums.ValueType), nullable=False))
    property_class: enums.PropertyClass = Field(sa_column=Column(Enum(enums.PropertyClass), nullable=False))
    unit: Optional[str] = Field(default=None)
    entity_type: enums.EntityType = Field(sa_column=Column(Enum(enums.EntityType), nullable=False))
    semantic_type_id: Optional[int] = Field(foreign_key=f"{DB_SCHEMA}.semantic_types.id", nullable=True, default=None)
    pattern: Optional[str] = Field(default=None)


class PropertyWithValue(PropertyBase):
    value_datetime: Optional[datetime] = None
    value_uuid: Optional[uuid.UUID] = None
    value_num: Optional[float] = None
    value_string: Optional[str] = None


class SynonymTypeBase(PropertyBase):
    """A specialized type of PropertyBase for synonyms with fixed constraints"""

    value_type: enums.ValueType = Field(default=enums.ValueType.string)
    property_class: enums.PropertyClass = Field(default=enums.PropertyClass.DECLARED)
    unit: Optional[str] = Field(default="")
    semantic_type_id: Optional[int] = Field(default=1)

    model_config = ConfigDict(populate_by_name=True)

    @field_validator("value_type")
    def validate_value_type(cls, v):
        if v != enums.ValueType.string:
            raise ValueError("SynonymType must have value_type of string")
        return v

    @field_validator("property_class")
    def validate_property_class(cls, v):
        if v != enums.PropertyClass.DECLARED:
            raise ValueError("SynonymType must have property_class of DECLARED")
        return v

    @field_validator("unit")
    def validate_unit(cls, v):
        if v not in (None, ""):
            raise ValueError("SynonymType must have empty unit")
        return v


class PropertyResponse(PropertyBase):
    id: int = Field(primary_key=True, index=True)
    created_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now()))


class AssayProperty(SQLModel, table=True):
    __tablename__ = "assay_properties"
    __table_args__ = {"schema": DB_SCHEMA}

    assay_id: int = Field(foreign_key=f"{DB_SCHEMA}.assays.id", primary_key=True)
    property_id: int = Field(foreign_key=f"{DB_SCHEMA}.properties.id", primary_key=True)
    required: bool = Field(default=False)

    assay: "Assay" = Relationship(back_populates="property_requirements")
    property: "Property" = Relationship()


class AssayBase(SQLModel):
    name: str = Field(nullable=False)
    description: Optional[str] = Field(default=None)


class AssayCreate(AssayBase):
    property_ids: List[int] = []  # List of property IDs to associate with this assay type
    property_requirements: List[Dict[str, Any]] = []  # List of property requirements
    property_details: List[Dict[str, Any]] = []  # List of property metadata


class AssayResponseBase(AssayBase):
    id: int = Field(primary_key=True, index=True)
    created_at: datetime = Field(sa_column=Column(DateTime, server_default=func.now()))
    updated_at: datetime = Field(sa_column=Column(DateTime, server_default=func.now(), onupdate=func.now()))


class AssayResponse(AssayResponseBase):
    properties: List["Property"] = []
    assay_details: List["AssayDetail"] = []
    property_requirements: List["AssayProperty"] = []


class Assay(AssayResponseBase, table=True):
    __tablename__ = "assays"
    __table_args__ = {"schema": DB_SCHEMA}

    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)

    properties: List["Property"] = Relationship(
        back_populates="assays", link_model=AssayProperty, sa_relationship_kwargs={"viewonly": True}
    )

    assay_runs: List["AssayRun"] = Relationship(back_populates="assay")
    assay_details: List["AssayDetail"] = Relationship(back_populates="assay")
    property_requirements: List["AssayProperty"] = Relationship(back_populates="assay")


class AssayRunBase(SQLModel):
    name: str = Field(nullable=False)
    description: Optional[str] = Field(default=None)
    assay_id: int = Field(foreign_key=f"{DB_SCHEMA}.assays.id", nullable=False)


class AssayRunResponseBase(AssayRunBase):
    id: int = Field(primary_key=True, index=True)
    created_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now()))


class AssayRunResponse(AssayRunResponseBase):
    assay: Optional["Assay"] = None
    assay_run_details: List["AssayRunDetail"] = []
    properties: List["Property"] = []


class AssayRunCreate(AssayRunBase):
    property_ids: List[int] = []  # List of property IDs to associate with this assay


class AssayRunDetail(SQLModel, table=True):
    __tablename__ = "assay_run_details"
    __table_args__ = {"schema": DB_SCHEMA}

    assay_run_id: int = Field(foreign_key=f"{DB_SCHEMA}.assay_runs.id", primary_key=True)
    property_id: int = Field(foreign_key=f"{DB_SCHEMA}.properties.id", primary_key=True)

    value_datetime: Optional[datetime] = Field(default=None, sa_column=Column(DateTime(timezone=True)))
    value_uuid: Optional[uuid.UUID] = Field(default_factory=uuid.uuid4)
    value_num: Optional[float] = Field(default=None)
    value_string: Optional[str] = Field(default=None)

    assay_run: "AssayRun" = Relationship(back_populates="assay_run_details")
    property: Optional["Property"] = Relationship()


class AssayRun(AssayRunResponseBase, table=True):
    __tablename__ = "assay_runs"
    __table_args__ = {"schema": DB_SCHEMA}

    updated_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now()))
    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)

    # Relationships - use assay_properties via assay to get list of expected properties
    # No direct properties relationship as assay_properties table no longer exists
    assay: "Assay" = Relationship(back_populates="assay_runs")
    assay_results: List["AssayResult"] = Relationship(back_populates="assay_run")
    assay_run_details: List["AssayRunDetail"] = Relationship(back_populates="assay_run")

    properties: List["Property"] = Relationship(
        back_populates="assay_runs",
        link_model=AssayRunDetail,
        sa_relationship_kwargs={"lazy": "joined", "viewonly": True},
    )


class AssayResultDetail(SQLModel, table=True):
    __tablename__ = "assay_result_details"
    __table_args__ = {"schema": DB_SCHEMA}

    assay_result_id: int = Field(foreign_key=f"{DB_SCHEMA}.assay_results.id", primary_key=True)
    property_id: int = Field(foreign_key=f"{DB_SCHEMA}.properties.id", primary_key=True)

    value_qualifier: Optional[int] = Field(default=0, nullable=False)  # 0 for "=", 1 for "<", 2 for ">"
    value_num: Optional[float] = Field(default=None)
    value_string: Optional[str] = Field(default=None)
    value_bool: Optional[bool] = Field(default=None)

    assay_result: "AssayResult" = Relationship(back_populates="assay_result_details")
    property: Optional["Property"] = Relationship()


class Property(PropertyResponse, table=True):
    __tablename__ = "properties"
    __table_args__ = (
        CheckConstraint(
            "value_type IN ('int', 'double', 'bool', 'datetime', 'string', 'uuid')", name="properties_value_type_check"
        ),
        CheckConstraint(
            "property_class IN ('DECLARED', 'CALCULATED', 'MEASURED', 'PREDICTED')",
            name="properties_property_class_check",
        ),
        CheckConstraint(
            "entity_type IN ('BATCH', 'COMPOUND', 'ASSAY', 'ASSAY_TYPES', 'ASSAY_RESULT', 'SYSTEM')",
            name="properties_entity_type_check",
        ),
        {"schema": DB_SCHEMA},
    )

    updated_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now()))
    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)

    batch_details: List["BatchDetail"] = Relationship(back_populates="property")
    compound_details: List["CompoundDetail"] = Relationship(back_populates="property")

    assays: List["Assay"] = Relationship(link_model=AssayProperty, sa_relationship_kwargs={"viewonly": True})
    assay_runs: List["AssayRun"] = Relationship(
        back_populates="properties", link_model=AssayRunDetail, sa_relationship_kwargs={"viewonly": True}
    )
    assay_results: List["AssayResult"] = Relationship(
        back_populates="properties", link_model=AssayResultDetail, sa_relationship_kwargs={"viewonly": True}
    )
    compounds: List["Compound"] = Relationship(
        back_populates="properties", link_model=CompoundDetail, sa_relationship_kwargs={"viewonly": True}
    )
    batches: List["Batch"] = Relationship(
        back_populates="properties", link_model=BatchDetail, sa_relationship_kwargs={"viewonly": True}
    )


class AssayResultBase(SQLModel):
    batch_id: int = Field(foreign_key=f"{DB_SCHEMA}.batches.id", nullable=False)
    assay_run_id: int = Field(foreign_key=f"{DB_SCHEMA}.assay_runs.id", nullable=False)
    updated_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now()))
    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)


class AssayResultResponseBase(AssayResultBase):
    id: int = Field(primary_key=True, index=True)


# Extended response model with backward compatibility field
class AssayResultResponse(AssayResultResponseBase):
    # Add a computed field for backward compatibility
    result_value: Optional[Union[float, str, bool]] = None

    @field_validator("result_value")
    def compute_result_value(cls, v, values):
        """Compute result_value from the appropriate typed value field"""
        if "value_num" in values and values["value_num"] is not None:
            return values["value_num"]
        elif "value_string" in values and values["value_string"] is not None:
            return values["value_string"]
        elif "value_bool" in values and values["value_bool"] is not None:
            return values["value_bool"]
        return None


class AssayResult(AssayResultResponseBase, table=True):
    __tablename__ = "assay_results"
    __table_args__ = {"schema": DB_SCHEMA}

    batch: "Batch" = Relationship(back_populates="assay_results")
    assay_run: "AssayRun" = Relationship(back_populates="assay_results")
    assay_result_details: List["AssayResultDetail"] = Relationship(back_populates="assay_result")

    properties: List["Property"] = Relationship(
        back_populates="assay_results",
        link_model=AssayResultDetail,
        sa_relationship_kwargs={"lazy": "joined", "viewonly": True},
    )


class AssayDetail(SQLModel, table=True):
    __tablename__ = "assay_details"
    __table_args__ = {"schema": DB_SCHEMA}

    assay_id: int = Field(foreign_key=f"{DB_SCHEMA}.assays.id", primary_key=True)
    property_id: int = Field(foreign_key=f"{DB_SCHEMA}.properties.id", primary_key=True)

    value_datetime: Optional[datetime] = Field(default=None, sa_column=Column(DateTime(timezone=True)))
    value_uuid: Optional[uuid.UUID] = Field(default_factory=uuid.uuid4)
    value_num: Optional[float] = Field(default=None)
    value_string: Optional[str] = Field(default=None)

    assay: "Assay" = Relationship(back_populates="assay_details")
    property: "Property" = Relationship()


class BatchAssayResultsCreate(SQLModel):
    assay_run_id: int
    batch_id: int
    measurements: Dict[str, Union[float, str, bool, Dict[str, Any]]]


class BatchAssayResultsResponse(SQLModel):
    assay_run_id: int
    batch_id: int
    assay_name: str
    measurements: Dict[str, Union[float, str, bool, Dict[str, Any]]]


class AdditionFields(SQLModel):
    name: Optional[str] = Field(default=None, unique=True)
    description: Optional[str] = None
    code: Optional[str] = None
    smiles: Optional[str] = None
    role: Optional[enums.AdditionsRole] = Field(default=None, sa_column=Column(Enum(enums.AdditionsRole)))
    molfile: Optional[str] = None
    formula: Optional[str] = None
    # Issue related to alias not working properly in SQLModel: https://github.com/fastapi/sqlmodel/issues/374
    molecular_weight: Optional[float] = Field(
        default=None,
        schema_extra={"validation_alias": "molecular weight"},
    )

    model_config = ConfigDict(populate_by_name=True)


class AdditionBase(AdditionFields):
    name: str = Field(nullable=False, unique=True)
    role: enums.AdditionsRole = Field(sa_column=Column(Enum(enums.AdditionsRole)))


class AdditionUpdate(AdditionFields):
    pass


class Addition(AdditionBase, table=True):
    __tablename__ = "additions"
    __table_args__ = (
        CheckConstraint("role IN ('SALT', 'SOLVATE')", name="additions_role_check"),
        {"schema": DB_SCHEMA},
    )

    id: int = Field(primary_key=True, index=True)
    created_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now(), nullable=False))
    updated_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now(), onupdate=func.now(), nullable=False)
    )
    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    is_active: bool = Field(default=True)
    is_archived: bool = Field(default=False)
    deleted_at: Optional[datetime] = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now()))
    deleted_by: Optional[uuid.UUID] = Field(default=None)


class BatchAdditionBase(SQLModel):
    batch_id: int = Field(foreign_key="moltrack.batches.id", nullable=False, unique=True)
    addition_id: int = Field(foreign_key="moltrack.additions.id", nullable=False, unique=True)
    addition_equivalent: float = Field(default=1)


class BatchAddition(BatchAdditionBase, table=True):
    __tablename__ = "batch_additions"
    __table_args__ = {"schema": DB_SCHEMA}

    id: int = Field(primary_key=True, index=True)
    created_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now(), nullable=False))
    updated_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now(), onupdate=func.now(), nullable=False)
    )
    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)

    batch: "Batch" = Relationship(back_populates="batch_additions")


class SchemaPayload(SQLModel):
    properties: List["PropertyBase"] = Field(default_factory=list)
    synonym_types: List["SynonymTypeBase"] = Field(default_factory=list)


class AssayPropertiesPayload(SQLModel):
    assay_type_details_properties: List["PropertyBase"] = Field(default_factory=list)
    assay_details_properties: List["PropertyBase"] = Field(default_factory=list)
    assay_type_properties: List["PropertyBase"] = Field(default_factory=list)


class AdditionsPayload(SQLModel):
    additions: List["AdditionBase"] = Field(default_factory=list)


class SchemaBatchResponse(SQLModel):
    properties: Optional[List["PropertyBase"]] = Field(default_factory=list)
    synonym_types: Optional[List["SynonymTypeBase"]] = Field(default_factory=list)
    additions: List["AdditionBase"] = Field(default_factory=list)


class AssayTypeCreateBase(SQLModel):
    name: str
    description: Optional[str] = None
    details: Dict[str, Any]


class AssayTypeCreate(SQLModel):
    assay_type: AssayTypeCreateBase


class AssayResultProperty(SQLModel):
    name: str
    required: bool


class AssayCreateBase(AssayBase):
    assay_result_properties: List[AssayResultProperty]
    extra_fields: Dict[str, Any] = Field(default_factory=dict)

    model_config = {"extra": "allow"}

    @model_validator(mode="before")
    def collect_extra_fields(cls, values):
        known_keys = {"name", "assay_result_properties"}
        extra = {k: v for k, v in values.items() if k not in known_keys}
        values["extra_fields"] = extra
        return values


class UpdateCheckResult(NamedTuple):
    action: str
    update_data: Optional[Dict[str, Any]] = None


# Advanced Search Models - New Recursive Structure
class AtomicCondition(SQLModel):
    """Individual atomic search condition with field, operator, and value"""

    field: str  # e.g., "compounds.canonical_smiles", "compounds.details.chembl"
    operator: enums.CompareOp
    value: Any
    threshold: Optional[float] = None  # For similarity searches (e.g., molecular similarity)

    @field_validator("field")
    def validate_field(cls, v):
        # Basic field format validation
        if not v or not isinstance(v, str):
            raise ValueError("Field must be a non-empty string")

        # Check for valid field format (table.field or table.details.property)
        parts = v.split(".")
        if len(parts) < 2 or len(parts) > 3:
            raise ValueError("Field must be in format 'table.field' or 'table.details.property'")

        valid_tables = get_args(Level)
        if parts[0] not in valid_tables:
            allowed = ", ".join(valid_tables)
            raise ValueError(f"Invalid table: {parts[0]}. Must be one of {allowed}")

        return v

    @model_validator(mode="before")
    def validate_threshold(cls, values):
        # Validate threshold is provided for operators that require it
        if isinstance(values, AtomicCondition):
            operator = values.get("operator")
            threshold = values.get("threshold")
            if operator == enums.CompareOp.IS_SIMILAR:
                if threshold is None:
                    raise ValueError(f"Operator {operator.value} requires a threshold value")
            elif threshold is not None:
                raise ValueError(f"Threshold not supported for operator: {operator}")
        return values


class LogicalNode(SQLModel):
    """Logical node combining multiple filters with AND/OR operator"""

    operator: enums.LogicOp
    conditions: List["Filter"]

    @field_validator("conditions")
    def validate_conditions(cls, v):
        if not v or len(v) == 0:
            raise ValueError("LogicalNode must have at least one condition")
        if len(v) == 1:
            raise ValueError("LogicalNode with single condition is redundant, use AtomicCondition directly")
        return v


# Filter is a union type - can be either AtomicCondition or LogicalNode
Filter = Union[AtomicCondition, LogicalNode]


Level = Literal["compounds", "batches", "assay_results", "assay_runs", "assays"]


class SearchRequest(SQLModel):
    """Main search request model with recursive filter structure"""

    level: Level
    output: List[str]  # Columns to return
    filter: Optional[Filter] = None
    output_format: enums.SearchOutputFormat

    @field_validator("output")
    def validate_output(cls, v):
        if not v:
            raise ValueError("Output must specify at least one column")
        return v


class SearchResponse(SQLModel):
    """Search response model"""

    status: str
    data: List[Dict[str, Any]]
    total_count: int
    level: str
    columns: List[str]


# Update forward references for recursive types
LogicalNode.model_rebuild()


class Token(NamedTuple):
    type: str
    value: Union[str, float, bool, None]



from sqlalchemy.orm import Session
from fastapi import HTTPException
from typing import List, Optional
from app.crud.properties import bulk_create_if_not_exists
from app import models

from app.utils import enums
from rdkit import Chem
from rdkit.Chem import Descriptors, rdMolDescriptors

from datetime import datetime


def enrich_addition(add: models.AdditionBase) -> models.AdditionBase:
    smiles, molfile, formula, mw = add.smiles, add.molfile, add.formula, add.molecular_weight

    mol = Chem.MolFromSmiles(smiles) if smiles else None
    if not mol and molfile:
        try:
            mol = Chem.MolFromMolBlock(molfile, sanitize=True)
        except Exception:
            mol = None

    if not mol:
        return add

    add.smiles = smiles or Chem.MolToSmiles(mol)
    add.molfile = molfile or Chem.MolToMolBlock(mol)
    add.formula = formula or rdMolDescriptors.CalcMolFormula(mol)
    add.molecular_weight = mw or Descriptors.MolWt(mol)

    return add


def create_additions(db: Session, additions: list[models.AdditionBase]) -> list[dict]:
    enriched_additions = [enrich_addition(add) for add in additions]
    return bulk_create_if_not_exists(db, models.Addition, models.AdditionBase, enriched_additions, validate=False)


def get_additions(db: Session, role: enums.AdditionsRole | None = None) -> List[models.AdditionBase]:
    query = db.query(models.Addition)
    if role is None:
        return query.all()
    return query.filter_by(role=role).all()


def get_addition_by_id(db: Session, addition_id: int) -> models.Addition:
    db_addition = db.get(models.Addition, addition_id)
    if db_addition is None:
        raise HTTPException(status_code=404, detail="Addition not found")
    return db_addition


def get_batch_addition_for_addition(db: Session, addition_id: int) -> Optional[models.BatchAddition]:
    return db.query(models.BatchAddition).filter(models.BatchAddition.addition_id == addition_id).first()


def update_addition_by_id(db: Session, addition_id: int, addition_update: models.AdditionUpdate):
    db_addition = get_addition_by_id(db, addition_id=addition_id)
    update_data = addition_update.dict(exclude_unset=True)
    for key, value in update_data.items():
        setattr(db_addition, key, value)

    db_addition.updated_at = datetime.now()
    db.add(db_addition)
    db.commit()
    db.refresh(db_addition)
    return db_addition


def delete_addition_by_id(db: Session, addition_id: int):
    db_addition = get_addition_by_id(db, addition_id=addition_id)
    db.delete(db_addition)
    db.commit()
    return db_addition
